import sys

if sys.version_info[:2] >= (3, 8):
    # TODO: Import directly (no need for conditional) when `python_requires = >= 3.8`
    from importlib.metadata import PackageNotFoundError, version  # pragma: no cover
else:
    from importlib_metadata import PackageNotFoundError, version  # pragma: no cover

try:
    # Change here if project is renamed and does not equal the package name
    dist_name = "contexere"
    __version__ = version(dist_name)
except PackageNotFoundError:  # pragma: no cover
    __version__ = "unknown"
finally:
    del version, PackageNotFoundError

__month_dict__ = dict([(m, abbr) for m, abbr in zip(range(1, 13),
                                               map(chr, range(ord('o'), ord('z') + 1))
                                               )])
__days__ = list(map(str, range(1, 10))) + list(map(chr, range(ord('A'), ord('V') + 1)))
__day_dict__ = dict([(d, abbr) for d, abbr in zip(range(1, 32), __days__)])

__hours__ = list(map(str, range(1, 1))) + list(map(chr, range(ord('a'), ord('x') + 1)))



import argparse
import logging
from pathlib import Path
import sys

from contexere import __version__
from contexere.discover import summary
from contexere.scheme import abbreviate_date, abbreviate_time, suggest_next

__author__ = "Andreas W. Kempa-Liehr"
__copyright__ = "Andreas W. Kempa-Liehr"
__license__ = "MIT"

_logger = logging.getLogger(__name__)


def parse_args(args):
    """Parse command line parameters

    Args:
      args (List[str]): command line parameters as list of strings
          (for example  ``["--help"]``).

    Returns:
      :obj:`argparse.Namespace`: command line parameters namespace
    """
    parser = argparse.ArgumentParser(description="Suggest name for research artefact")
    parser.add_argument(
        "--version",
        action="version",
        version=f"contexere {__version__}",
    )
    parser.add_argument(dest="path",
                        help="Path to folder with research artefacts (default: current working dir)",
                        nargs='?',
                        type=Path,
                        default=Path.cwd())
    parser.add_argument(
        "-n",
        "--next",
        dest="next",
        help="Suggest next artefact name",
        action="store_true"
    )
    parser.add_argument(
        "-p",
        "--project",
        dest="project",
        type=str,
        default=None,
        help="Specify project abbreviation",
        action="store"
    )
    parser.add_argument(
        "-s",
        "--summary",
        dest="summary",
        help="Sumarize files following the naming convention",
        action="store_true"
    )
    parser.add_argument(
        "-t",
        "--time",
        dest="time",
        help="add time abbreviation",
        action="store_true"
    )  
    parser.add_argument(
        "-u",
        "--utc",
        dest="utc",
        help="Generate timestamp with respect to UTC (default is local timezone)",
        action="store_true"
    )
    parser.add_argument(
        "-v",
        "--verbose",
        dest="loglevel",
        help="set loglevel to INFO",
        action="store_const",
        const=logging.INFO,
    )
    parser.add_argument(
        "-vv",
        "--very-verbose",
        dest="loglevel",
        help="set loglevel to DEBUG",
        action="store_const",
        const=logging.DEBUG,
    )
    return parser.parse_args(args)


def setup_logging(loglevel):
    """Setup basic logging

    Args:
      loglevel (int): minimum loglevel for emitting messages
    """
    logformat = "[%(asctime)s] %(levelname)s:%(name)s:%(message)s"
    logging.basicConfig(
        level=loglevel, stream=sys.stdout, format=logformat, datefmt="%Y-%m-%d %H:%M:%S"
    )


def main(args):
    args = parse_args(args)
    setup_logging(args.loglevel)
    _logger.debug("Start building context...")
    if args.summary:
        summary(args.path)
    elif args.next:
        print(suggest_next(args.path, project=args.project, local=~args.utc))
    else:
        output = abbreviate_date(local=~args.utc)
        if args.time:
            output += abbreviate_time(local=~args.utc)
        print(output)

    # print(args.project + abbreviate_date() + ending)
    _logger.info("Script ends here")


def run():
    """Calls :func:`main` passing the CLI arguments extracted from :obj:`sys.argv`

    This function can be used as entry point to create console scripts with setuptools.
    """
    main(sys.argv[1:])


if __name__ == "__main__":
    # ^  This is a guard statement that will prevent the following code from
    #    being executed in the case someone imports this file instead of
    #    executing it as a script.
    #    https://docs.python.org/3/library/__main__.html

    # After installing your project with pip, users can also run your Python
    # modules as scripts via the ``-m`` flag, as defined in PEP 338::
    #
    #     python -m contexere.name
    #
    run()



from datetime import datetime

from contexere.scheme import abbreviate_date

def test_first_day_of_month():
    day = datetime(2022, 1, 1)
    assert abbreviate_date(day) == '22o1'

def test_last_day_of_month_with_31_days():
    day = datetime(2022, 1, 31)
    assert abbreviate_date(day) == '22oV'

def test_last_day_of_February():
    day = datetime(2022, 2, 28)
    assert abbreviate_date(day) == '22pS'

def test_last_day_of_February_in_leap_year():
    day = datetime(2020, 2, 29)
    assert abbreviate_date(day) == '20pT'

def test_last_day_of_April():
    day = datetime(2022, 4, 30)
    assert abbreviate_date(day) == '22rU'

def test_number_character_transition():
    day = datetime(2022, 4, 9)
    assert abbreviate_date(day) == '22r9'

    day = datetime(2022, 4, 10)
    assert abbreviate_date(day) == '22rA'

def test_string_input():
    assert abbreviate_date('2022-04-03 23:00') == '22r3'



"""
Discover files following the naming convention
"""

import logging
import pandas as pd
from pathlib import Path
import re

# Define the scheme with named groups
pattern = re.compile(r'(?P<project>[a-zA-Z]*)(?P<date>[0-9]{2}[o-z][1-9A-V])(?P<step>[a-z]*)_')


# Function to group files by common project and date
def build_context(directory='.', project_filter=None):
    context = dict()
    timeline = dict()

    # Iterate over files and directories in the specified folder
    for path in Path(directory).iterdir():
        if path.is_file():
            match = pattern.match(path.name)
            if match:
                project = match.group('project')
                date = match.group('date')
                step = match.group('step')
                if project_filter is None or project == project_filter:
                    if not project in context:
                        context[project] = dict()
                    if not (date, step) in context[project]:
                        context[project][(date, step)] = list()
                    if not (date + step) in timeline:
                        timeline[date + step] = dict()
                    if not project in timeline[date + step]:
                        timeline[date + step][project] = list()
                    context[project][(date, step)].append(path)
                    timeline[date + step][project].append(path)
    return context, timeline
    
def last(timeline):
    events = list(timeline.keys())
    events.sort()
    latest = events[-1]
    return [project + latest for project in timeline[latest]]

def summary(directory='.'):
    context, timeline = build_context(directory)
    summary = pd.Series({project: len(context[project])
                         for project in context}).sort_values(ascending=False)

    if len(summary) == 0:
        raise ValueError('No context found in folder "{}".'.format(str(directory)))
    tail = last(timeline)
    artefacts = 'artefact' if len(tail) == 1 else 'artefacts'

    print(f"Summary of {'artefact' if len(summary) == 1 else 'artefacts'}:")
    print(summary)
    print(f"Last {'artefact' if len(tail) == 1 else 'artefacts'}: ", ', '.join(tail))

if __name__ == "__main__":
    summary()


import datetime
import logging
import pandas as pd
import pytz
import re
from tzlocal import get_localzone

from contexere import __month_dict__, __day_dict__, __hours__
from contexere.discover import build_context, last

# Define the scheme with named groups
schematic = re.compile(r'(?P<project>[a-zA-Z]*)(?P<date>[0-9]{2}[o-z][1-9A-V])(?P<step>[a-z]*)')


def abbreviate_date(date=None, tz=pytz.utc, local=False,
                    month=__month_dict__, day=__day_dict__):
    if date is None:
        if local:
            date = datetime.datetime.now(tz=get_localzone())
        else:
            date = datetime.datetime.now(tz=tz)
    elif type(date) == str:
        date = pd.Timestamp(date)
    year = date.strftime('%y')

    return year + month[date.month] + day[date.day]


def abbreviate_time(date=None, seconds=False, tz=pytz.utc,  local=False, hour=__hours__):
    if date is None:
        if local:
            date = datetime.datetime.now(tz=get_localzone())
        else:
            date = datetime.datetime.now(tz=tz)
    elif type(date) == str:
        date = pd.Timestamp(date)
    abbr = hour[date.hour] + '{:02}'.format(date.minute)
    if seconds:
        return abbr + '{:02}'.format(date.second)
    return abbr


def abbreviate_datetime(date=None, seconds=False, tz=pytz.utc):
    if date is None:
        date = datetime.datetime.now(tz=tz)
    return abbreviate_date(date) + abbreviate_time(date, seconds=seconds)


def decode_abbreviated_datetime(abrv, tz=pytz.utc):
    """
    Decode the 2021 naming scheme to a datetime object

    Args:
        abrv: String in format yymd[hMM]
              yy [0-9][0-9] encodes the years 2000 to 2099
              m [o-z] encodes the months with 'o' referring to January and
                                              'z' referring to December
              d [1-9,A-V] encodes the day, which is either the number, or
                                                            'A' referring to the 10th,
                                                        and 'V' referring to the 31st
              h [a-x] encodes the hour with 'a' referring to midnight and 'x' to 23
              MM [0-5][0-9] encodes the minutes 0 to 59
        tz: time zone info (default: pytz.utc)

    Returns: datetime object
    """
    assert len(abrv) == 4 or len(abrv) == 7
    year = int(abrv[:2]) + 2000
    month = ord(abrv[2]) - ord('o') + 1
    if abrv[3] in list(map(chr, range(ord('1'), ord('9') + 1))):
        day = int(abrv[3])
    else:
        day = ord(abrv[3]) - ord('A') + 10
    if len(abrv) == 7:
        hour = ord(abrv[4]) - ord('a')
        minutes = int(abrv[-2:])
    else:
        hour = 0
        minutes = 0
    return datetime.datetime(year, month, day, hour, minutes, tzinfo=tz)


def suggest_next(directory='.', project=None, local=True):
    context, timeline = build_context(directory, project_filter=project)
    logging.info('Projects' + str(list(context.keys())))
    logging.info('Timeline' + str(list(timeline.keys())))
    today = abbreviate_date(local=local)
    if len(timeline) == 0:
        if project is None:
            raise ValueError(f"No project files matching the naming scheme found in path {directory}"
                             "and option '--project' wasn't set.")
        else:
            this_project = project
            next_step = 'a'
    else:
        latest = last(timeline)
        match = schematic.match(latest[0])
        this_project = match.group('project')
        if today == match.group('date'):
            assert match.group('step') != 'z'
            next_step = chr(ord(match.group('step')) + 1)
        else:
            next_step = 'a'
    suggestion = this_project + today + next_step
    return suggestion


from datetime import datetime
import pytz

from contexere.scheme import decode_abbreviated_datetime

def test_day_only():
    assert decode_abbreviated_datetime('22pM') == datetime(2022, 2, 22, tzinfo=pytz.timezone('UTC'))

def test_datetime():
    assert decode_abbreviated_datetime('22oJf00') == datetime(2022, 1, 19, 5, tzinfo=pytz.timezone('UTC'))

def test_datetime_with_seconds():
    assert decode_abbreviated_datetime('22oJf53') == datetime(2022, 1, 19, 5, 53, tzinfo=pytz.timezone('UTC'))



import pandas as pd

from contexere.scheme import abbreviate_date, abbreviate_time, decode_abbreviated_datetime


class Calendar(object):
    def __init__(self):
        self.cache_ = dict()

    def __call__(self, dt=None):
        if dt is None:
            abbreviated_datetime = abbreviate_date() + abbreviate_time()
        else:
            abbreviated_datetime = abbreviate_date(dt) + abbreviate_time(dt)
        self.cache_[abbreviated_datetime] = dt
        return abbreviated_datetime

    def __getitem__(self, abbreviated_datetime):
        if not abbreviated_datetime in self.cache_.keys():
            self.cache_[abbreviated_datetime] = decode_abbreviated_datetime(abbreviated_datetime)
        return self.cache_[abbreviated_datetime]

    def forward(self, start, horizon):
        """
        Compute abbreviated time stamp in the future

        >>> calendar = Calendar()
        >>> calendar.forward('21vMu00', pd.Timedelta('12h'))
        '21vNi00'

        Args:
            start: Abbreviated datetime string
            horizon: pd.Timedelta object (default pd.Timedelta('12h'))

        Returns:
            abbreviated datetime
        """
        start_dt = self.__getitem__(start)
        end_dt = start_dt + horizon
        return abbreviate_date(end_dt) + abbreviate_time(end_dt)

    def backward(self, start, horizon):
        """
        Compute abbreviated time stamp in the past

        >>> calendar = Calendar()
        >>> calendar.backward('21vMu00', pd.Timedelta('12h'))
        '21vMi00'

        Args:
            start: Abbreviated datetime string
            horizon: pd.Timedelta object (default pd.Timedelta('12h'))

        Returns:
            abbreviated datetime
        """
        start_dt = self.__getitem__(start)
        end_dt = start_dt - horizon
        return abbreviate_date(end_dt) + abbreviate_time(end_dt)

    def in_seconds(self, abbreviated_datetime):
        return self.__getitem__(abbreviated_datetime).timestamp()




class DailyEnumerator(object):
    """
    Iterator for counters of daily data sets.
    """

    def __init__(self, N=None, previous=None):
        if previous is not None:
            for letter in previous:
                assert letter >= 'a' and letter <= 'z'
        self.start_value = 'a'
        self.previous = previous
        self.N = N

    def __iter__(self):
        self.value = self.previous
        self.n = 0
        return self

    def __next__(self):
        if self.n == self.N:
            raise StopIteration
        self.n += 1

        if self.value is None:
            self.value = self.start_value
        else:
            self.value = self.plus_one(self.value)
        return self.value

    def plus_one(self, stump):
        if len(stump) == 0:
            return 'a'
        elif stump[-1] < 'z':
            return stump[:-1] + chr(ord(stump[-1]) + 1)
        else:
            return self.plus_one(stump[:-1]) + 'a'


from datetime import datetime
import pytz

from contexere.logbook import Calendar

def test_day_only():
    calendar = Calendar()
    assert calendar['22pM'] == datetime(2022, 2, 22, tzinfo=pytz.timezone('UTC'))

def test_datetime():
    calendar = Calendar()
    assert calendar['22oJf00'] == datetime(2022, 1, 19, 5, tzinfo=pytz.timezone('UTC'))

def test_datetime_with_seconds():
    calendar = Calendar()
    assert calendar['22oJf53'] == datetime(2022, 1, 19, 5, 53, tzinfo=pytz.timezone('UTC'))

def test_midnight():
    calendar = Calendar()

    dt = datetime(2022, 1, 1, 0, 0)
    assert calendar(dt) == '22o1a00'

    dt = datetime(2022, 1, 1, 1, 42)
    assert calendar(dt) == '22o1b42'

def test_arbitrary_time():
    calendar = Calendar()
    dt = datetime(2022, 1, 19, 5, 54)

    assert calendar(dt) == '22oJf54'

def test_time_zone_agnostic():
    calendar = Calendar()
    dt = datetime(2022, 1, 19, 5, 54, tzinfo=pytz.timezone("Pacific/Auckland"))

    assert calendar(dt) == '22oJf54'


from datetime import datetime
import pytz

from contexere.scheme import abbreviate_time

def test_midnight():
    dt = datetime(2022, 1, 1, 0, 0)
    assert abbreviate_time(dt) == 'a00'
    assert abbreviate_time(dt, seconds=True) == 'a0000'

    dt = datetime(2022, 1, 1, 1, 42)
    assert abbreviate_time(dt) == 'b42'
    assert abbreviate_time(dt, seconds=True) == 'b4200'

def test_arbitrary_time():
    dt = datetime(2022, 1, 19, 5, 54)
    assert abbreviate_time(dt) == 'f54'

def test_time_zone_agnostic():
    dt = datetime(2022, 1, 19, 5, 54, tzinfo=pytz.timezone("Pacific/Auckland"))
    assert abbreviate_time(dt) == 'f54'

def test_string_input():
    assert abbreviate_time('2022-04-03 23:00') == 'x00'
